#include <zephyr/types.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/kernel.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>

#include "../Vulcan_UUID/Vulcan_Uuid.h"
#include "../Vulcan_nvs/nvsStorage.h"
#include "../Vulcan_RingState/RingState.h"

#define VERSION_ID 					1
#define INIT_NAME_ID 				2
#define INIT_HARDWARE_ID 			3

#define UPDATE_START_MSG    "START_OTA"
#define UPDATE_END_MSG      "END_OTA"

#define VERSION             		"0.7"
#define INIT_NAME           		"RingSonDinh"
#define INIT_HARDWARE				"RING V2"

#define LOCK      					"140499"
#define LOCK_LEN  					6

#define MAX_VERSION_LENGTH  	10
#define MAX_NAME_LENGTH     	30
#define MAX_HARDWARE_LENGTH		30
#define DATA_LEN 25

uint8_t ota_value[DATA_LEN + 1];
uint8_t name_value[DATA_LEN + 1];
uint8_t hardware_value[DATA_LEN + 1];
uint8_t countControl_value[DATA_LEN + 1];

char name[MAX_NAME_LENGTH + 1];
char hardware[MAX_HARDWARE_LENGTH + 1];
char version[MAX_VERSION_LENGTH + 1];

ssize_t onReadOTA(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			void *buf, uint16_t len, uint16_t offset)
{
	char *value = attr->user_data;
	read_Init_DataString_Device_Version(VERSION_ID, VERSION, version, sizeof(version));
	memcpy(value, version, strlen(version));
	//__NVIC_SystemReset(); api reset chip
	
	return bt_gatt_attr_read(conn, attr, buf, len, offset, value,
				 strlen(value));
}

ssize_t onReadName(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			void *buf, uint16_t len, uint16_t offset)
{
	char *value = attr->user_data;
	memcpy(value, name, strlen(name));

	return bt_gatt_attr_read(conn, attr, buf, len, offset, value,
				 strlen(value));
}

ssize_t onWriteName(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			 const void *buf, uint16_t len, uint16_t offset,
			 uint8_t flags)
{
	uint8_t *value = attr->user_data;

	char s[len+1];
	memset(s, 0 , sizeof(s));
	memcpy(s, buf, len);
	if (strlen(s) > 0) {
		memcpy(name, s, strlen(s) + 1);
		save_nvs_String(INIT_NAME_ID, name);
		printk("Mame: %s\n", name);
		notifyConfirm(true);
	}else {
		notifyConfirm(false);
	}

	if (offset + len > DATA_LEN) {
		return BT_GATT_ERR(BT_ATT_ERR_INVALID_OFFSET);
	}
	memcpy(value, buf, len);
	value[offset + len] = 0;
	
	return len;
}

ssize_t onReadHardware(struct bt_conn *conn, const struct bt_gatt_attr *attr, 
			void *buf, uint16_t len, uint16_t offset)
{
	char *value = attr->user_data;
	memcpy(value, hardware, strlen(hardware));

	return bt_gatt_attr_read(conn, attr, buf, len, offset, value,
				 strlen(value));
}

ssize_t onWriteHardware(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			 const void *buf, uint16_t len, uint16_t offset,
			 uint8_t flags)
{
	uint8_t *value = attr->user_data;

	char s[len+1];
	memset(s, 0, sizeof(s));
	memcpy(s, buf, len);

	if (strlen(s) > LOCK_LEN) {
		char key[LOCK_LEN + 1];
		strncpy(key, s, LOCK_LEN);
		key[LOCK_LEN] = '\0';

		if (strcmp(key, LOCK) == 0) {
			printk("key is valid\n");
			memcpy(hardware, s + LOCK_LEN, strlen(s) - LOCK_LEN);
			printk("key hardware: %s\n", hardware);
			save_nvs_String(INIT_HARDWARE_ID, hardware);
			notifyConfirm(true);
		}
		else {
			printk("key is unvalid");
			notifyConfirm(false);
		}
	}
	else {
		notifyConfirm(false);
	}

	if (offset + len > DATA_LEN) {
		return BT_GATT_ERR(BT_ATT_ERR_INVALID_OFFSET);
	}
	memcpy(value + offset, buf, len);
	value[offset + len] = 0;
	
	return len;
}

ssize_t onReadCountControl(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			void *buf, uint16_t len, uint16_t offset)
{
	char *value = attr->user_data;

	return bt_gatt_attr_read(conn, attr, buf, len, offset, value,
				 strlen(value));
}

ssize_t onWriteCountControl(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			 const void *buf, uint16_t len, uint16_t offset,
			 uint8_t flags)
{
	uint8_t *value = (uint8_t *)attr->user_data;
	
	if (offset + len > DATA_LEN) {
		return BT_GATT_ERR(BT_ATT_ERR_INVALID_OFFSET);
	}
	memcpy(value + offset, buf, len);
        printk("%s\n", value);
	value[offset + len] = 0;
	
	return len;
}

BT_GATT_SERVICE_DEFINE(ringService,
	BT_GATT_PRIMARY_SERVICE(&ring_Service_UUID),
	BT_GATT_CHARACTERISTIC(&ota_UUID.uuid,
			       BT_GATT_CHRC_READ ,
			       BT_GATT_PERM_READ | BT_GATT_PERM_WRITE,
			       onReadOTA, NULL, ota_value),
	BT_GATT_CHARACTERISTIC(&name_Char_UUID.uuid,
			       BT_GATT_CHRC_READ | BT_GATT_CHRC_WRITE,
			       BT_GATT_PERM_READ | BT_GATT_PERM_WRITE,
			       onReadName, onWriteName, name_value),
	BT_GATT_CHARACTERISTIC(&hardware_Char_UUID.uuid,
			       BT_GATT_CHRC_READ | BT_GATT_CHRC_WRITE,
			       BT_GATT_PERM_READ | BT_GATT_PERM_WRITE,
			       onReadHardware, onWriteHardware, hardware_value),
	BT_GATT_CHARACTERISTIC(&count_Control_UUID.uuid,
			       BT_GATT_CHRC_READ | BT_GATT_CHRC_WRITE,
			       BT_GATT_PERM_READ | BT_GATT_PERM_WRITE,
			       onReadCountControl, onWriteCountControl, countControl_value),
);

void readDataRing(void) {
	read_Init_DataString_Device_Version(VERSION_ID, VERSION, version, sizeof(version));
	read_Init_DataString_Device(INIT_NAME_ID, INIT_NAME, name, sizeof(name));
	read_Init_DataString_Device(INIT_HARDWARE_ID, INIT_HARDWARE, hardware, sizeof(hardware));	
}

const char* getName() {
	return name;
}