#include <inttypes.h>
#include <stddef.h>
#include <stdint.h>

#include <zephyr/types.h>
#include <string.h>
#include <errno.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/drivers/adc.h>
#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/util.h>

#include <zephyr/kernel.h>
#include <zephyr/drivers/gpio.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>

#include <stdio.h>
#include <stdlib.h>

#include <zephyr/bluetooth/hci_vs.h>
#include <zephyr/drivers/bluetooth/hci_driver.h>

#include "../Vulcan_UUID/Vulcan_Uuid.h"
#include "../Vulcan_BLERing/BLEEMGSensor.h"
#include "../Vulcan_BLERing/BLERing.h"
#include "../WS2812/ws2812.h"

uint8_t countDeviceConnected = 0;

#define TIME_CONNECT   60000
bool stateConnect = false;

#define MAX_NAME_LENGTH     11
char nowNameBLE[31];

// #define DEVICE_NAME CONFIG_BT_DEVICE_NAME
char DEVICE_NAME[MAX_NAME_LENGTH + 1];
#define DEVICE_NAME_LEN (sizeof(DEVICE_NAME) - 1)

struct bt_le_adv_param adv;

static uint8_t mfg_dataStatic[8] = {0xff, 0xff};
// static uint8_t iphone[] = {0x06, 0x00, 0x12, 0x00};

static const struct bt_data ad[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
	BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, 12),
	BT_DATA(BT_DATA_SVC_DATA16, mfg_dataStatic, 8),
	// BT_DATA(BT_DATA_PERIPHERAL_INT_RANGE, iphone, 4),
};

static const struct bt_data sd[] = {
	BT_DATA_BYTES(BT_DATA_UUID128_ALL, BT_UUID_CUSTOM_SERVICE_MYOBAND),
};

void mtu_updated(struct bt_conn *conn, uint16_t tx, uint16_t rx)
{
	printk("Updated MTU: TX: %d RX: %d bytes\n", tx, rx);
}

static struct bt_gatt_cb gatt_callbacks = {
	.att_mtu_updated = mtu_updated
};

void connected(struct bt_conn *conn, uint8_t err)
{
	if (err) {
		printk("Connection failed (err 0x%02x)\n", err);
	} else {
		stateConnect = true;
		countDeviceConnected++;
		printk("Connected\n");
	}
}

void disconnected(struct bt_conn *conn, uint8_t reason)
{
	stateConnect = false;
	countDeviceConnected--;
	printk("Disconnected (reason 0x%02x)\n", reason);
}

static bool le_param_req(struct bt_conn *conn, struct bt_le_conn_param *param)
{
	printk("LE conn  param req: int (0x%04x, 0x%04x) lat %d"
		    " to %d\n", param->interval_min, param->interval_max,
		    param->latency, param->timeout);

	return true;
}

void le_param_updated(struct bt_conn *conn, uint16_t interval,
			     uint16_t latency, uint16_t timeout)
{
	printk("LE conn param updated: int 0x%04x lat %d "
		     "to %d\n", interval, latency, timeout);
}

BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected = connected,
	.disconnected = disconnected,
	.le_param_req = le_param_req,
	.le_param_updated = le_param_updated,
};

int settings_load(void);

void setupAdvertising() {

	char nameTemp[MAX_NAME_LENGTH + 1];
	memset(nameTemp, 0, sizeof(nameTemp));
	strncpy(nameTemp, getName(), MAX_NAME_LENGTH);
	memset(DEVICE_NAME, 0 ,sizeof(DEVICE_NAME));
	memcpy(DEVICE_NAME, nameTemp, strlen(nameTemp));
	memset(nowNameBLE, 0, sizeof(nowNameBLE));
	memcpy(nowNameBLE, getName(), strlen(getName()));
	bt_set_name(DEVICE_NAME);
	
	adv.id = BT_ID_DEFAULT;
	adv.interval_max = BT_GAP_ADV_FAST_INT_MAX_2;
	adv.interval_min = BT_GAP_ADV_FAST_INT_MIN_2;
	adv.options = BT_LE_ADV_OPT_CONNECTABLE;
	adv.peer = NULL;
	adv.secondary_max_skip = 0;
	adv.sid = 0;
}

void startAdvertising() {
	char *newNameBLE = getName();
	if (strcmp(newNameBLE, nowNameBLE) != 0) {
		setupAdvertising();
		printk("Name had been changed\n");
		bt_le_adv_stop();
		int err = bt_le_adv_start(&adv, ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
		if (err) {
			printk("Advertising failed to start (err %d)\n", err);
			return;
		}
		printk("Advertising successfully started\n");	
	}
	
}

void stopAdvertising() {
	bt_le_adv_stop();
}

uint8_t getCountDevice() {
	return countDeviceConnected;
}

void bt_ready(void)
{
	int err;

	printk("Bluetooth initialized\n");

	if (IS_ENABLED(CONFIG_SETTINGS)) {
		settings_load();
	}

	setupAdvertising();
	err = bt_le_adv_start(&adv, ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
	if (err) {
		printk("Advertising failed to start (err %d)\n", err);
		return;
	}

	printk("Advertising successfully started\n");
        
}

void begin_bt(void) {
    // Enable BLE 
	int err = bt_enable(NULL);
	if (err) {
		printk("Bluetooth init failed (err %d)\n", err);
	}

	static char addr_str[BT_ADDR_STR_LEN];
	struct bt_hci_vs_static_addr addr;
	int addr_count;

	addr_count = bt_read_static_addr(&addr, 1);

	if (addr_count > 0)
	{
		bt_addr_to_str(&(addr.bdaddr), addr_str, sizeof(addr_str));
		printk("BT addr: %s\n", addr_str);
	}

	for (uint8_t i = 5, j = 2; i >= 0 && j < 8; i--, j++)
	{
		mfg_dataStatic[j] = addr.bdaddr.val[i];
	}

	for (uint8_t i = 0; i < 8; i++)
	{
		printk("%x\t", mfg_dataStatic[i]);
	}
	printk("\n");

	bt_ready();
	bt_gatt_cb_register(&gatt_callbacks);
}

void read_Data_Information_Device(void) {
	readDataRing();
}

bool waitConnect(const struct gpio_dt_spec *specPinTurnOff, uint16_t _checkTimeWakeup) {
	uint32_t lastTime = k_uptime_get_32();
	while ((k_uptime_get_32() - lastTime < TIME_CONNECT) && (!stateConnect)) {
		whiteLight();
		k_msleep(300);
		clearLight();
		k_msleep(1000);

		uint32_t timeWakeUp;
		if (gpio_pin_get_dt(specPinTurnOff))
		{
			timeWakeUp = k_uptime_get_32();
			k_msleep(100);
		}

		while (gpio_pin_get_dt(specPinTurnOff))
		{
			if (k_uptime_get_32() - timeWakeUp > _checkTimeWakeup)
			{
				redLight();
				while (gpio_pin_get_dt(specPinTurnOff));
				return stateConnect;
				k_msleep(100);
			}
		}
	}

	if (stateConnect) {
		greenLight();
		k_msleep(200);
		clearLight();
		k_msleep(500);
		greenLight();
		k_msleep(200);
	}
	return stateConnect;
}

void notifyData(uint8_t _state) {
	onNotifySignal(_state);
}

void sendControl(int8_t stateControl) {
	onNotifyStateControl(stateControl);
}

void notifyBatery(uint8_t battery) {
	onNotifyBattery(battery);
}

uint8_t getModeReadSensorManager() {
	return getModeReadSensor();
}