#include <inttypes.h>
#include <stddef.h>
#include <stdint.h>
#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>

#include <inttypes.h>
#include <stddef.h>
#include <stdint.h>

#include <zephyr/types.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/kernel.h>

#include <zephyr/devicetree.h>
#include <zephyr/sys/util.h>
#include <zephyr/drivers/spi.h>

#include <zephyr/kernel.h>
#include <zephyr/sys/reboot.h>
#include <zephyr/device.h>
#include <string.h>
#include <zephyr/drivers/flash.h>
#include <zephyr/storage/flash_map.h>
#include <zephyr/fs/nvs.h>

#include "../Vulcan_nvs/nvsStorage.h"
#include "../WS2812/ws2812.h"
#include "../MCP3208/mcp3208.h"
#include "../Vulcan_Filter/Vulcan_Utils.h"
#include "../Vulcan_FFT/fft.h"

#define SPIBB_NODE	DT_NODELABEL(spibb0)

#define THRESHOLD_SETUP 			60
#define THRESHOLD_DC    			40
#define THRESHOLD_LOW   			10
#define THRESHOLD_HIGH  			20
#define THRESHOLD_MAX   			1000

#define TIME_READY 						3000
#define TIME_SETUP_SENSOR			30000

#define ADC_VREF    					3300    				// 3.3V Vref
#define CHANELS     					3      					// samples

#define SIZE_WRITE						CHANELS * 4

#define SIZE_ONE_PACKET_REV		sizeof(float) * (CHANELS + 1) // array CHANELS + 1 float (CHANELS sensor + 1 time read)
#define NUMBER_PACKET_REV			10							//10 packet/send
#define LENGTH_DATA_REV				NUMBER_PACKET_REV * SIZE_ONE_PACKET_REV // <600 byte

#define DATA_CALIB_LINE_MAX									50
#define SIZE_ONE_PACKET_CALIB_UINT16				5 	//THRESHOLD_SETUP | THRESHOLD_LOW | THRESHOLD_HIGH | TIME
#define SIZE_DATA_SAVE_CALIB_SD_UINT16			SIZE_ONE_PACKET_CALIB_UINT16 * DATA_CALIB_LINE_MAX			

#define SIZE_ONE_PACKET_CALIB_BYTE					SIZE_ONE_PACKET_CALIB_UINT16 * sizeof(uint16_t)
#define SIZE_DATA_SAVE_CALIB_SD_BYTE				SIZE_ONE_PACKET_CALIB_BYTE * DATA_CALIB_LINE_MAX

#define CALIB_KEY							"threshold"

#define FILE_DATA							"/data.txt"
#define FILE_CALIB						"/calib.txt"

#define THRESHOLD_ID	4
#define MODE_LOGIC_ID	5

uint16_t buffDataCalibSD[SIZE_DATA_SAVE_CALIB_SD_UINT16];

static uint8_t buffer[SIZE_ONE_PACKET_REV];
uint8_t rev[LENGTH_DATA_REV];
static float data[CHANELS + 1] = {0}; // 8 chanels + timeCycle
float dataFilter[CHANELS] = {0};

uint16_t SWSPL_FREQ;

float ftHighSensor[8];
float ftLowSensor[8];
float ftLowSensor1[8];
float fs = 230;

// State Ring
bool stateReady = false;

// Variable of logic Control
uint8_t modeLogic = 1;
int8_t open = -1;
int8_t close = 1;

uint8_t stateControlEMG = 0;

// Variable of time read
int8_t stateSensorEMG = 0;

uint32_t timeCycle;
uint32_t timeSetRead;
// uint32_t timeRead = micros();
uint8_t chanels = 3; // number of channels

uint32_t _timeReady;

// Dynamic memory pointer
float *dataPoint;

// Variable to store sensor signal using flash memory
//  File file;
bool isFirtRead = true;
bool isStartSend = true;

uint32_t count = 0;
uint32_t lengthData = 0;

// Threshold for control processing
uint16_t threshold[3] = {THRESHOLD_SETUP, THRESHOLD_LOW, THRESHOLD_HIGH};
uint16_t test = 15;

// Variable for sensor signal processing
uint16_t sensor[CHANELS];
float value[CHANELS];
float valueDC[CHANELS];
float valueOut[CHANELS];
float filterKalman[CHANELS];
float output;
float outputDC[3];

bool stateWait = false;

uint16_t getVref() {
	return ADC_VREF;
}

float getSampleRate() {
	return 1000000.0/timeCycle;
}

void setSampleRate(uint16_t sampleRate) {
	SWSPL_FREQ = sampleRate;
}

void restoreThreshold(void) {
	// threshold[3]= {THRESHOLD_SETUP, THRESHOLD_LOW, THRESHOLD_HIGH};
	threshold[0] = THRESHOLD_SETUP;
	threshold[1] = THRESHOLD_LOW;
	threshold[2] = THRESHOLD_HIGH;
}

void readSensor() {
    sensor[0] = readADC(0);
	sensor[1] = readADC(5);
	sensor[2] = readADC(7);
    // printk("%d\t%d\t%d\n", sensor[0], sensor[1], sensor[2]);
}

void filterSensor() {
	for (uint8_t i = 0; i < CHANELS; i++) {
		UTILS_LOW_PASS_FILTER(value[i], abs(highpassB(sensor[i], i)), 0.05);
	}
	// printk("%.2f\t%.2f\t%.2f\n", value[0], value[1], value[2]);
}

void setSensorDC() {
	for(uint8_t i = 0; i < CHANELS; i++) {
		valueDC[i] = lowpassB(value[i], i);
	}
	// printk("%.2f\t%.2f\t%.2f\n", value[0], value[1], value[2]);
}

void smoothSensor() {
	for(uint8_t i = 0; i < CHANELS; i++) {
		valueOut[i] = value[i] - valueDC[i];
		// valueOut[i] = ftLowSensor1[i].lowpass(value[i]);
		UTILS_LOW_PASS_FILTER(data[i], valueOut[i], 0.03);
	}
	// printk("%.2f\t%.2f\t%.2f\n", data[0], data[1], data[2]);
}

void signalProcessing() {
	float output = data[0] + data[1] + data[2];
	stateControlEMG = output > threshold[2] ? close 
								:output < threshold[1] ? open : 0;
}

int8_t getStateControl() {
	return stateControlEMG;
}

uint8_t readSensorSDBLE() {
	// stateSensor:
	// 0: don't read
	// 1: firstRead
	// 2: onRead
	// 3: endRead

	// _state return:
	// 0: don't read
	// 1: read raw signal
	// 2: read filter signal

	// if (stateSensor == 0) {
	// 	return 0;
	// }
	// if (isFirtRead) {
	// 	// setupData();
	// 	setupSD();
	// 	isFirtRead = false;
	// 	count = 0;
	// 	Serial.println("Begin read sensor");
	// 	timeRead = micros();
	// 	return 1;
	// }
	// if (micros() - timeRead >= timeSetRead) {
	// 	// closeData();
	// 	closeSD();
	// 	stateSensor = 0;
	// 	Serial.println("End read sensor");
	// 	Serial.printf("Total number of samples: %d\n", count);
	// 	Serial.printf("Sampling frequency: %d\n",count/(timeSetRead/1000000));
	// 	return 3;
	// }
	// // static uint32_t lasttime = micros();
	// // if (micros() - lasttime >= 4000) {
	// 	// lasttime = micros();
	// 	readSensor();
	// 	filterSensor();
	// 	smoothSensor();
	// 	// writeData();
	// 	writeSD();
	// 	count++;
	// 	// Serial.println(count++);
		
	// // }
	return 2;
}
uint16_t countest = 0;
uint8_t readSensorStreamBLE() {
	// stateSensor:
	// 0: don't read
	// !0: onRead

	// _state return:
	// 0: don't read
	// 2: onRead
	if (stateSensorEMG == 0) {
		return 0;
	}
	readSensor();
	filterSensor();
	smoothSensor();
	memcpy((void *)buffer, (void *)data, SIZE_ONE_PACKET_REV);
	signalProcessing();

	return 2;
}

uint8_t readSensorBLE() {
	if (stateSensorEMG == 0) {
		return 0;
	}
	switch (stateSensorEMG) {
		case 1:
			return(readSensorSDBLE());
			break;
		case 2:
			return(readSensorStreamBLE());
			break;
		
		default:
			break;
	}
	return 0;
}

void setStateSensor(uint8_t _stateSensor) {
	isStartSend = true;
	stateSensorEMG = _stateSensor;
	isFirtRead = true;
}

void setTimeReadSensor(uint8_t _timeRead) {
	if (_timeRead == 0) {
		setStateSensor(0);
		printk("End Read Sensor\n");
		return;
	}
	if (_timeRead == UINT8_MAX) {
		setStateSensor(2);
		printk("Read sensor stream BLE\n");
	} else {
		timeSetRead = _timeRead * 1000 * 1000;
		printk("Set time read: %d s\n", _timeRead);
		setStateSensor(1);
	}
}

void setModeLogicControl(uint8_t _mode) {
	printk("Mode control: %d\n",_mode);
	if (_mode == 0) return;
	modeLogic = _mode;
	save_nvs_Int8(MODE_LOGIC_ID, &modeLogic);
	printk("Mode: %d\n", _mode);
	switch (_mode)
	{
		case 1:
			open = -1;
			close = 1;
			break;
		case 2: 
			open = 1;
			close = -1;
			break;
		default:
			break;
	}
}

bool isGetOnHand() {
	readSensor();
	filterSensor();
	if (value[0] < THRESHOLD_MAX && value[1] < THRESHOLD_MAX && value[2] < THRESHOLD_MAX)
	return true;
	else return false;
}

void firstReadSensor() {
	readSensor();
	filterSensor();
	setSensorDC();

	// for(uint8_t i = 0; i < CHANELS; i++) {
	// 	kalmanFt[i].setFirstValue(abs(ftHighSensor[i].highpass(data[i])));
	// }
}

bool waitSensor(uint16_t _threshold, uint16_t waitTime) {
	// static uint32_t _timeReady = k_uptime_get_32();
	readSensor();
	filterSensor();
	setSensorDC();
	// printk("%.2f\t%.2f\t%.2f\t\n", value[0], value[1], value[2]);
	// Serial.printf("%.2f\t%.2f\t%.2f\t", valueDC[0], valueDC[1], valueDC[2]);
	// printk();
	if ((value[0] < _threshold) && (value[1] < _threshold) && (value[2] < _threshold))
	{
		if (!stateWait)
		{
			_timeReady = k_uptime_get_32();
			stateWait = true;
		}
		if (k_uptime_get_32() - _timeReady > waitTime)
			return true;
		else
			return false;
	} else
	{
		stateWait = false;
		return false;
	}
}

bool waitRelax(uint16_t maxTimeWait) {
	stateReady = false;

	blueLight(); // Setup Ring LED BLUE
	memset(sensor, 0, sizeof(sensor));
	memset(value, 0, sizeof(value));
	memset(valueDC, 0, sizeof(valueDC));
	memset(valueOut, 0, sizeof(valueOut));
	memset(filterKalman, 0, sizeof(filterKalman));

	for (int i = 0; i < 5000; i++)
	{
		firstReadSensor();
	}
	printk("Myoband wait get on Hand\n");
	while (!isGetOnHand());

	printk("Myoband on Setup\n");

	stateWait = false;
	uint32_t lastTimeRelax = k_uptime_get_32();
	_timeReady = k_uptime_get_32();
	while (!waitSensor(threshold[0], TIME_READY))
	{
		if (maxTimeWait > 0) if (k_uptime_get_32() - lastTimeRelax > maxTimeWait) return false;
	}
	// printk("myoban ready");
	// printk("myoban wait relax");
	k_msleep(200);
	// stateWait = false;
	// while (!waitSensor(threshold[0], TIME_READY));
	printk("MyoBand running\n");
	k_msleep(200);
	greenLight(); // Ready Ring LED GREEN
	stateReady = true;
	return true;
}

bool begin_EMGSensor() {
    begin_KickFiltersRT();
	for (uint8_t i = 0; i < CHANELS; i++) {
		inithighpass(0, 5, fs, i);
		initlowpass(0, 0.1, fs, i);
	}
    // // nvs.begin_nvs();
	// // pref.begin("emgSensor", false);
	// // SPIFFS.begin(true);
    
    // // printk("State Check: %d\n", nvs.get_nvs(THRESHOLD_ID, threshold, sizeof(threshold), threshold));
    
	if (get_nvs_Threshold(THRESHOLD_ID, threshold, sizeof(threshold))) {
		printk("Threshold: %d, %d, %d\n", threshold[0], threshold[1], threshold[2]);
		if (threshold[0] == 0) {
			printk("Calibration is false.\nYou must be calibration again!!!");
			restoreThreshold();
		}
	} else restoreThreshold();
	printk("Threshold: %d, %d, %d\n", threshold[0], threshold[1], threshold[2]);

	setModeLogicControl(get_nvs_ModeLogic(MODE_LOGIC_ID, &modeLogic, sizeof(modeLogic)));
	printk("Setup Myo Band\n");
	// printCalibHistory();
	uint16_t tempTimeSetupSensor = TIME_SETUP_SENSOR;
	return waitRelax(tempTimeSetupSensor);
    // return 1;
}

int8_t syncEMGSensor() {
	if (!stateReady)
		return stateControlEMG = 0;
	readSensor();
	filterSensor();
	smoothSensor();
	float output = data[0] + data[1] + data[2];

	// printk("%.2f\t%.2f\t%.2f\t", data[0], data[1], data[2]);
	// printk("output: %f",output);
	// printk("\n");
	if (value[0] > THRESHOLD_MAX || value[1] > THRESHOLD_MAX || value[2] > THRESHOLD_MAX) {
		waitRelax(0);
		return stateControlEMG = 0;
	}
	else {
		if (output > threshold[2]) {
			// printk("Close\n");
			return stateControlEMG = close;
		}
		else {
			if (output < threshold[1]) {
				// printk("Open\n");
				return stateControlEMG = open;
			}
			else
				return stateControlEMG = 0;
		}
	}
}

uint8_t getModeLogicControl() {
	return modeLogic;
}

bool isReady() {
	return stateReady;
}

uint8_t *returnData() {
	return buffer;
}

uint8_t *returnRev() {
	return rev;
}

uint16_t *getThreshold() {
	int ret = get_nvs_Threshold(THRESHOLD_ID, threshold, sizeof(threshold));
	return threshold;
}

void setLed_EMG(uint8_t option) {
	//Setup Ring LED RED
	switch (option)
	{
		case 1:
			/* code */
			redLight();
			break;
		case 2:
			greenLight();
			break;
		case 3:
			blueLight();
			break;
		case 4:
			whiteLight();
			break;
		default:
			clearLight();
			break;
	}
}

void notifyConfirm_EMG(bool isSuccess) {
	uint32_t cLed = isSuccess 
									? 2 // Green Light
									: 1;// Red Light
									
	setLed_EMG(cLed);			//Setup Ring LED RED
	k_msleep(100);

	setLed_EMG(0);				// Clear Light
	k_msleep(100);

	setLed_EMG(cLed);			//Setup Ring LED RED
	k_msleep(100);

	setLed_EMG(0);				// Clear Light
	k_msleep(100);

	greenLight();			//Setup Ring LED RED
}

// Save data sensor by SD
void setupSD() {
	// file.close();
	// SPIFFS.remove(FILE_DATA);
	// file = SPIFFS.open(FILE_DATA, FILE_APPEND);
}

void writeSD() {
	memcpy((void *)buffer, (void *)data, SIZE_ONE_PACKET_REV);
	// file.write(buffer, SIZE_ONE_PACKET_REV);
}

void closeSD() {
	// printk("Size data: %d byte\n", file.size());
	// file.close();
	
}

void openSD() {
	// file.close();
	// file = SPIFFS.open(FILE_DATA);
}

int readSD() {
	if (isStartSend) {
		openSD();
		isStartSend = false;
		count = 0;
		return -2;
	}
	// if (file.available()) {
    // file.read(rev, LENGTH_DATA_REV);
	// 	for (uint8_t i = 0; i < NUMBER_PACKET_REV; i++) {
	// 		float buff[CHANELS + 1];
	// 		memcpy((void *)buff, (void *)(rev + SIZE_ONE_PACKET_REV * i), SIZE_ONE_PACKET_REV);
	// 		for (uint8_t j = 0; j < CHANELS + 1; j++) {
	// 			Serial.printf("%.2f\t", buff[j]);
	// 		}
	// 		Serial.print("\n");
	// 	}
	// 	count++;
	// 	return count;
	// }
	else {
		closeSD();
		isStartSend = true;
		return -1;
	}
}

void setupData() {
	clearData();
	dataPoint = (float *)malloc(sizeof(float) * 100000*(CHANELS + 1));
}

void writeData() {
	memcpy(&dataPoint[count * (CHANELS + 1)], (void*)data, sizeof(data));
}

void closeData() {
	lengthData = --count;
	for (uint16_t i = 0; i < lengthData; i++) {
		for (uint8_t j = 0; j < (CHANELS + 1); j++) {
			printk("%.2f \t", dataPoint[i * (CHANELS + 1) + j]);
		}
		printk("\n");
	}
}

void openData() {
	count = 0;
}

void clearData() {
	free(dataPoint);
}

int readData() {
	if (isStartSend) {
		openData();
		isStartSend = false;
		return -2;
	}
	if (count < lengthData) {
		uint16_t buff = lengthData - count;
		buff = (buff < NUMBER_PACKET_REV) ? buff:NUMBER_PACKET_REV;
		memcpy((void*)rev, &dataPoint[count], buff * SIZE_ONE_PACKET_REV);
		count += buff;
		return buff;
	}
	else {
		clearData();
		isStartSend = true;
		return -1;
	} 
}