#include <errno.h>
#include <string.h>
#include <zephyr/kernel.h>
#include <zephyr/drivers/led_strip.h>
#include <zephyr/device.h>
#include <zephyr/drivers/spi.h>
#include <zephyr/sys/util.h>
#include <zephyr/drivers/gpio.h>

#define STRIP_NODE		DT_ALIAS(led_strip)
#define STRIP_NUM_PIXELS	DT_PROP(DT_ALIAS(led_strip), chain_length)

#define DELAY_TIME K_MSEC(500)

#define RGB(_r, _g, _b) { .r = (_r), .g = (_g), .b = (_b) }

static const struct led_rgb colors[] = {
	RGB(0x00, 0x00, 0x00), /* off */
	RGB(0x5A, 0x00, 0x00), /* red */
	RGB(0x00, 0x5A, 0x00), /* green */
	RGB(0x00, 0x00, 0x5A), /* blue */
	RGB(0xA, 0x3C, 0x5A), /* white */
};

static struct led_rgb pixels[STRIP_NUM_PIXELS];

static const struct device *const strip = DEVICE_DT_GET(STRIP_NODE);

void init_ws2812(void) {
    if (device_is_ready(strip)) {
		printk("Found LED strip device %s\n", strip->name);
	} else {
		printk("LED strip device %s is not ready\n", strip->name);
	}
	printk("Displaying pattern on strip\n");
}

void clearLight(void) {
    memset(&pixels, 0x00, sizeof(pixels));
    memcpy(&pixels[0], &colors[0], sizeof(struct led_rgb));
    led_strip_update_rgb(strip, pixels, STRIP_NUM_PIXELS);
}

void redLight(void) {
    memset(&pixels, 0x00, sizeof(pixels));
    memcpy(&pixels[0], &colors[1], sizeof(struct led_rgb));
    led_strip_update_rgb(strip, pixels, STRIP_NUM_PIXELS);
}

void greenLight(void) {
    memset(&pixels, 0x00, sizeof(pixels));
    memcpy(&pixels[0], &colors[2], sizeof(struct led_rgb));
    led_strip_update_rgb(strip, pixels, STRIP_NUM_PIXELS);
}

void blueLight(void) {
    memset(&pixels, 0x00, sizeof(pixels));
    memcpy(&pixels[0], &colors[3], sizeof(struct led_rgb));
    led_strip_update_rgb(strip, pixels, STRIP_NUM_PIXELS);
}
 
void whiteLight() {
    memset(&pixels, 0x00, sizeof(pixels));
    memcpy(&pixels[0], &colors[4], sizeof(struct led_rgb));
    led_strip_update_rgb(strip, pixels, STRIP_NUM_PIXELS);
}