#include <zephyr/kernel.h>
#include <inttypes.h>
#include <stddef.h>
#include <stdint.h>

#include <zephyr/types.h>
#include <string.h>
#include <errno.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/drivers/adc.h>
#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/util.h>
#include <zephyr/drivers/spi.h>

#include <zephyr/kernel.h>
#include <zephyr/drivers/gpio.h>

#include <stdio.h>
#include <stdlib.h>

#include <zephyr/sys/reboot.h>
#include <zephyr/drivers/flash.h>
#include <zephyr/storage/flash_map.h>
#include <zephyr/fs/nvs.h>

#include <zephyr/init.h>
#include <zephyr/pm/pm.h>
#include <zephyr/pm/device.h>
#include <zephyr/pm/policy.h>
#include <soc.h>
#include <hal/nrf_gpio.h>

#include <zephyr/mgmt/mcumgr/transport/smp_bt.h>

#include "WS2812/ws2812.h"
#include "Vulcan_Adc/PrefAdc.h"
#include "Vulcan_Filter/Vulcan_Utils.h"
#include "Vulcan_nvs/nvsStorage.h"
#include "MCP3208/mcp3208.h"
#include "Vulcan_BLERing/BLERingManager.h"
#include "Vulcan_EMGSensor/EMGSensor.h"
#include "Vulcan_RingState/RingState.h"

// #define DEBUG

#define READ_SENSOR_OP_STACK_SIZE           5000
#define READ_SENSOR_BLE_OP_STACK_SIZE       5000
#define MODE_OP_STACK_SIZE                  5000
#define RING_OP_STACK_SIZE                  5000

#define READ_SENSOR_OP_PRIORITY           1
#define READ_SENSOR_BLE_OP_PRIORITY       1
#define MODE_OP_PRIORITY                  2
#define RING_OP_PRIORITY                  2

#define TIME_WAKEUP   1500
#define BATTERY_MAX   2360
#define BATTERY_MIN   1820

// /*
//                           Read Adc
//   0 --> 3.6V  |   4.15V phan ap -> 2.075V   |  3.2 phan ap --> 1.6V
//   0 --> 4095  |                    2360     |                  1820 
// */

K_THREAD_STACK_DEFINE(stack_ReadSensorOP, READ_SENSOR_OP_STACK_SIZE);
K_THREAD_STACK_DEFINE(stack_ReadSensorBLEOP, READ_SENSOR_BLE_OP_STACK_SIZE);
K_THREAD_STACK_DEFINE(stack_ModeOP, MODE_OP_STACK_SIZE);
K_THREAD_STACK_DEFINE(stack_RingOP, RING_OP_STACK_SIZE);

struct k_thread addressReadSensorOP;
struct k_thread addressReadSensorBLEOP;
struct k_thread addressModeOP;
struct k_thread addressRingOP;

extern void ReadSensorOP(void *, void *, void *);
extern void ReadSensorBLEOP(void *, void *, void *);
extern void ModeOP(void *, void *, void *);
extern void RingOP(void *, void *, void *);

#define SPIBB_NODE	DT_NODELABEL(spibb0)
#define LEDEN_NODE DT_ALIAS(leden)
static const struct gpio_dt_spec ledEn = GPIO_DT_SPEC_GET(LEDEN_NODE, gpios);
#define LDO_NODE DT_ALIAS(pinldo)
static const struct gpio_dt_spec ldo = GPIO_DT_SPEC_GET(LDO_NODE, gpios);
#define SWRING_NODE	DT_ALIAS(swring)
#if !DT_NODE_HAS_STATUS(SWRING_NODE, okay)
#error "Unsupported board: sw0 devicetree alias is not defined"
#endif
static const struct gpio_dt_spec button = GPIO_DT_SPEC_GET_OR(SWRING_NODE, gpios,
							      {1});

//Create object FS 
static struct nvs_fs fs;

uint8_t stateSensor = 0;
uint8_t mode = 0;
int8_t stateControl = 0;
bool stateSleep = true;

uint8_t connectedCountBLE = 0;
uint32_t lastTimeBattery;

void deepSleep(void);
void onSleep(bool _stateSleep);
void onWakeUp(void);
void onPower(void);
void onCharger(void);
uint8_t getBattery(void);

void ReadSensorOP(void *, void *, void *) {
  k_thread_suspend(&addressReadSensorOP);
  for (;;) {
    // printk("ReadSensor\n");
    stateControl = syncEMGSensor();
    k_usleep(10);
  }
}

void ReadSensorBLEOP(void *, void *, void *) {
    const struct device *const dev = DEVICE_DT_GET(SPIBB_NODE);

    if (!device_is_ready(dev)) {
        printk("%s: device not ready.\n", dev->name);
    }

    struct spi_cs_control cs_ctrl = (struct spi_cs_control){
        .gpio = GPIO_DT_SPEC_GET(SPIBB_NODE, cs_gpios),
        .delay = 0u,
    };
    begin_MCP3208(dev, &cs_ctrl);
    k_thread_suspend(&addressReadSensorBLEOP);
    for (;;) {
        // printk("ReadSensorBLE\n");
        notifyData(readSensorBLE());
        sendControl(getStateControl());
        k_usleep(10);
    }
}

void ModeOP(void *, void *, void *) {
  int8_t lastState = 0;
  uint32_t lastTime = k_uptime_get_32();
  k_thread_resume(&addressModeOP);
  notifyBatery(getBattery());
  for(;;) {
    // printk("ModeOP\n");
    int8_t _state = sync_Ring(stateControl);
    if (lastState != _state) {
      // Serial.println(_state);
      sendControl(_state);
      lastState = _state;
    }
    if (k_uptime_get_32() - lastTime > 3000) {
    //   printk("Battery Test: %d\n", getBattery());
      notifyBatery(getBattery());
      lastTime = k_uptime_get_32();
    }
    k_msleep(1);
  }
}

void RingOP(void *, void *, void *) {
    lastTimeBattery = k_uptime_get_32();
    for (;;) {
        // printk("RingOP\n");
        connectedCountBLE = getCountDevice();
        // printk("Connect Device: %d\n", getCountDevice());
        if ((connectedCountBLE != 0)) {
            stateSensor = getModeReadSensorManager();
            if (stateSensor == 0)
            {
                k_thread_suspend(&addressReadSensorBLEOP);
                k_thread_resume(&addressModeOP);
                k_thread_resume(&addressReadSensorOP);
            }
            else
            {
                k_thread_suspend(&addressModeOP);
                k_thread_suspend(&addressReadSensorOP);
                k_thread_resume(&addressReadSensorBLEOP);
            }
        }
        onPower();
        k_msleep(100);
    }
}

int main() {
    printk("This is a Vulcan Augmetics's product: Ring RayTak V2.1\n");
    int ret;
    if (!gpio_is_ready_dt(&button)) {
		printk("Error: button device %s is not ready\n",
		       button.port->name);
		return 0;
	}
	ret = gpio_pin_configure(button.port, (gpio_port_pins_t)button.pin, GPIO_INPUT);
	if (ret != 0) {
		printk("Error %d: failed to configure %s pin %d\n",
		       ret, button.port->name, button.pin);
		return 0;
	}

    /* Configure to generate PORT event (wakeup) on button 1 press. */
	nrf_gpio_cfg_input(NRF_DT_GPIOS_TO_PSEL(DT_ALIAS(swring), gpios),
			   NRF_GPIO_PIN_PULLDOWN);
	nrf_gpio_cfg_sense_set(NRF_DT_GPIOS_TO_PSEL(DT_ALIAS(swring), gpios),
			       NRF_GPIO_PIN_SENSE_HIGH);

    if (!gpio_is_ready_dt(&ledEn)) {
        return 0;
    }

    if (!gpio_is_ready_dt(&ldo)) {
        return 0;
    }

    ret = gpio_pin_configure_dt(&ledEn, GPIO_OUTPUT_LOW);
    if (ret < 0) {
        return 0;
    }

    ret = gpio_pin_configure_dt(&ldo, GPIO_OUTPUT_LOW);
    if (ret < 0) {
        return 0;
    }

    const struct device *const dev = DEVICE_DT_GET(SPIBB_NODE);

    if (!device_is_ready(dev))
    {
        printk("%s: device not ready.\n", dev->name);
        return 0;
    }

    struct spi_cs_control cs_ctrl = (struct spi_cs_control){
        .gpio = GPIO_DT_SPEC_GET(SPIBB_NODE, cs_gpios),
        .delay = 0u,
    };

    gpio_pin_set_raw(ldo.port, ldo.pin, 1);
    gpio_pin_set_raw(ledEn.port, ledEn.pin, 1);

    // stateSleep = false;
    printk("State sleep: %d\n",stateSleep);
    onWakeUp();
    
    begin_Adc();
    init_ws2812();
    begin_NVS(&fs);
    read_Data_Information_Device();

    begin_MCP3208(dev, &cs_ctrl);

    begin_RingState();
    stateSleep = !begin_EMGSensor();
    onSleep(stateSleep);
    begin_bt();     //Enable BLE
    stateSleep = !waitConnect(&button, TIME_WAKEUP);
    onSleep(stateSleep);
    
    printk("Start\n");
    k_thread_create(&addressReadSensorOP, stack_ReadSensorOP, K_THREAD_STACK_SIZEOF(stack_ReadSensorOP), ReadSensorOP, NULL, NULL, NULL, READ_SENSOR_OP_PRIORITY, 0, K_NO_WAIT);
    k_thread_create(&addressReadSensorBLEOP, stack_ReadSensorBLEOP, K_THREAD_STACK_SIZEOF(stack_ReadSensorBLEOP), ReadSensorBLEOP, NULL, NULL, NULL, READ_SENSOR_BLE_OP_PRIORITY, 0, K_NO_WAIT);
    k_thread_create(&addressModeOP, stack_ModeOP, K_THREAD_STACK_SIZEOF(stack_ModeOP), ModeOP, NULL, NULL, NULL, MODE_OP_PRIORITY, 0, K_NO_WAIT);
    k_thread_create(&addressRingOP, stack_RingOP, K_THREAD_STACK_SIZEOF(stack_RingOP), RingOP, NULL, NULL, NULL, RING_OP_PRIORITY, 0, K_NO_WAIT);

    return 0;
}

/// @brief 
/// @param 
void onSleep(bool _stateSleep) {
  if (!_stateSleep) return;
  printk("On deep sleep\n");
  gpio_pin_set_raw(ledEn.port, ledEn.pin, 1);
  redLight();
  k_msleep(250);
  clearLight();
  k_msleep(250);

  redLight();
  k_msleep(250);
  clearLight(); 

  stopAdvertising();
  gpio_pin_set_raw(ldo.port, ldo.pin, 0);
  gpio_pin_set_raw(ledEn.port, ledEn.pin, 0); 
  
  deepSleep();
}

void onWakeUp() {
//    #ifndef DEBUG
//     while (!gpio_pin_get_dt(&button)) {
//       onCharger();
//       k_msleep(1000);
//       __NVIC_SystemReset(); api reset chip
//     }
//   #endif

    uint32_t timeWakeUp = 0;
    if (gpio_pin_get_dt(&button)) {
        timeWakeUp = k_uptime_get_32();
        k_msleep(100);
    }
    while (gpio_pin_get_dt(&button)) {
        if (k_uptime_get_32() - timeWakeUp > TIME_WAKEUP) {
            stateSleep = !stateSleep;
            printk("State sleep: %d\n",stateSleep);
            redLight();
            while(gpio_pin_get_dt(&button)) {
                k_msleep(10);
            }
            k_msleep(100);
        }
    }
    onSleep(stateSleep);
}

void onPower()
{
    // #ifndef DEBUG
    //   if (!gpio_pin_get_dt(&button)) __NVIC_SystemReset(); api reset chip
    // #endif

    uint32_t timeWakeUp = 0;
    if (gpio_pin_get_dt(&button))
    {
        timeWakeUp = k_uptime_get_32();
        k_msleep(100);
    }

    while (gpio_pin_get_dt(&button))
    {
        if (k_uptime_get_32() - timeWakeUp > TIME_WAKEUP)
        {
            stateSleep = !stateSleep;
            printk("State sleep: %d\n", stateSleep);
            redLight();
            while (gpio_pin_get_dt(&button))
            {
                k_msleep(10);
            }
            k_msleep(100);
        }
    }

    if (!stateSleep) {
        switch (connectedCountBLE)
        {
            case 0:
                if (!isReady()) {
                    stateSleep = true;
                    onSleep(stateSleep);
                }
                startAdvertising();
                stateSleep = !waitConnect(&button, TIME_WAKEUP);
                if (!stateSleep) {
                    if (isReady())
                    {
                        greenLight();
                    }
                }
                break;
            case 1:
                startAdvertising();
                break;

        default:
            stopAdvertising();
            break;
        }
    }

    if (k_uptime_get_32() - lastTimeBattery > 1000) {
        uint8_t batteryPersent = getBattery();
        if (batteryPersent < 15) {
            for (uint8_t i = 0; i < 3; i++) {
                redLight();
                k_msleep(200);

                redLight();
                k_msleep(200);
            }
            if (batteryPersent < 5) {
                redLight();
                k_msleep(2000);
                stateSleep = true;
            }
        }
        lastTimeBattery = k_uptime_get_32();
    }

    onSleep(stateSleep);
}

void onCharger() {
  uint8_t batteryPercent = getBattery();
  // printk("Battery Percent");
  // printk("\t");
  // printk("%d\n",batteryPercent);
  if (batteryPercent > 95)      greenLight();
  else                          redLight();
}

uint8_t getBattery() {
  float Vin = analogRead(0);   //resistor voltage divider
  UTILS_LOW_PASS_FILTER(Vin, analogRead(0), 0.01);
  int16_t batteryPercent = (Vin - BATTERY_MIN) / (BATTERY_MAX - BATTERY_MIN) * 100.0;
  batteryPercent = constrain(batteryPercent, 0, 100);
  return batteryPercent;
}

void deepSleep(void) {
    printk("System off Ring\n");
    struct pm_state_info info;
    info.state = PM_STATE_SOFT_OFF;
    info.min_residency_us = 0;
    info.exit_latency_us = 0;
    pm_state_force(0u, &info);
}